local addonName, addon = ...

-- functions
local format = string.format
local ipairs = ipairs
local pairs = pairs
local tinsert = table.insert
local tsort = table.sort
-- constants
-- frames
-- libs
local LibStub = LibStub
local LibQTip = LibStub("LibQTip-1.0")

local private = {
    broker = nil,
    optionName = "useClassicTooltip",

    compiled = false,
    totalKeystones = 0,
    totalBestRuns = 0,
    totalPreviousBestRuns = 0,
    realms = nil,
}

local function ResetTooltipInformation()
    addon:Debug("Resetting tooltip information.")

    private.compiled = false
    private.totalKeystones = 0
    private.totalBestRuns = 0
    private.totalPreviousBestRuns = 0
    private.realms = nil
end

local function CompileTooltipInformation()
    addon:Debug("Compiling tooltip information.")

    local totalKeystones = 0
    local totalBestRuns = 0
    local totalPreviousBestRuns = 0

    local characters = addon:GetCharacters(function(_, c)
        if c.keystone then totalKeystones = totalKeystones + 1 end
        if c.record then totalBestRuns = totalBestRuns + 1 end
        if c.previousBestRun then totalPreviousBestRuns = totalPreviousBestRuns + 1 end

        return c.keystone ~= nil or c.record ~= nil or c.previousBestRun ~= nil
    end)

    local thisRealm = addon:GetRealmName()
    local realms = {}
    do
        local characterKey = addon:GetCharacterKey()
        local keys = {}
        for _, v in pairs(characters) do
            local key = keys[v.realm] or (#realms + 1)
            keys[v.realm] = key
            realms[key] = realms[key] or {
                name = v.realm,
                numKeystones = 0,
                characters = {},
            }

            if v.keystone then
                realms[key].numKeystones = realms[key].numKeystones + 1
            end

            tinsert(realms[key].characters, v)

            tsort(realms[key].characters, function(a, b)
                local aKey = addon:GetCharacterKey(a)
                if aKey == characterKey then
                    return true
                end

                local bKey = addon:GetCharacterKey(b)
                if bKey == characterKey then
                    return false
                end

                return a.name < b.name
            end)

        end
        tsort(realms, function(a, b)
            return a == thisRealm or a.name < b.name
        end)
    end

    private.compiled = true
    private.totalKeystones = totalKeystones
    private.totalBestRuns = totalBestRuns
    private.totalPreviousBestRuns = totalPreviousBestRuns
    private.realms = realms
end

local function plural(num, singular, pluralised)
    return tonumber(num or 0) == 1 and singular or (pluralised or format("%ss", singular))
end

local function OnTooltipShow(tooltip)
    local totalKeystones = private.totalKeystones
    local totalBestRuns = private.totalBestRuns
    local totalPreviousBestRuns = private.totalPreviousBestRuns
    local hasWeeklyBestStaleDataBug = false

    if totalPreviousBestRuns > 0 then
        local row = tooltip:AddLine()
        tooltip:SetCell(row, 1, format(
            "You have an unopened Mythic+ reward chest on |cffFFCC66%d|r %s; visit your %s to loot it.",
            totalPreviousBestRuns,
            plural(totalPreviousBestRuns, "character"),
            addon.isBFA and "faction hub" or "class order hall"
        ), nil, nil, 3, nil, nil, nil, 300)

        tooltip:AddSeparator()
    end

    if totalKeystones < 1 then
        local row = tooltip:AddLine()
        tooltip:SetCell(row, 1, "You have no keystones.  Run a dungeon on mythic difficulty or above to receive one!", nil, nil, addon.options.showRecords and 3 or 2, nil, nil, nil, 300)

        if addon.options.showRecords and (totalBestRuns > 0 or totalPreviousBestRuns > 0) then
            tooltip:AddSeparator()
        end
    end

    if totalKeystones < 1 and (not addon.options.showRecords or (totalBestRuns < 1 and totalPreviousBestRuns < 1)) then
        return
    end

    do
        local row = tooltip:AddHeader()
        tooltip:SetCell(row, 1, "Character")
        tooltip:SetCell(row, 2, "Keystone")
        if addon.options.showRecords then
            tooltip:SetCell(row, 3, "Best Run")
        end
    end

    for _, realm in ipairs(private.realms) do
        tooltip:AddSeparator()
        do
            local row = tooltip:AddHeader()
            tooltip:SetCell(row, 1, format("|cffFFCC66%d|r %s", realm.numKeystones, realm.name), nil, nil, addon.options.showRecords and 3 or 2)
            tooltip:AddSeparator()
        end

        for _, character in ipairs(realm.characters) do
            local key = addon:GetCharacterKey(character)
            local row = tooltip:AddLine()

            if character.previousBestRun then
                tooltip:SetCell(row, 1, format("|cffFF0000**|r%s", addon:GetCharacterDisplayName(key, true, true)))
            else
                tooltip:SetCell(row, 1, addon:GetCharacterDisplayName(key, true, true))
            end

            if character.keystone then
                tooltip:SetCell(row, 2, addon:GetKeystoneDisplayText(character.keystone, true, true))
            end

            if (character.record or character.previousBestRun) and addon.options.showRecords then
                if character.record[1] == 0 then
                    hasWeeklyBestStaleDataBug = true
                end

                if character.previousBestRun then
                    tooltip:SetCell(row, 3, format("|cff666666%s|r", addon:GetRecordDisplayText(character.previousBestRun)))
                else
                    tooltip:SetCell(row, 3, addon:GetRecordDisplayText(character.record, true))
                end
            end

            if addon.options.highlightCharactersWithAvailableRewards and character.previousBestRun then
                tooltip:SetLineColor(row, 1, 0, 0, 0.5)
            end
        end
    end

    if hasWeeklyBestStaleDataBug then
        tooltip:AddSeparator()
        do
            local row = tooltip:AddHeader()
            tooltip:SetCell(row, 1, format([[Please note: some of your best run dungeons|nare displayed as just "Mythic+" due a bug with|n8.0 not returning accurate dungeon completion|ninformation.]]), nil, nil, addon.options.showRecords and 3 or 2)
        end
    end

    -- tooltip:AddSeparator()
    -- do
    --     local row = tooltip:AddLine()
    --     tooltip:SetCell(row, 1, format("|cffFFCC66%d|r total", totalNumKeystones), nil, "LEFT")
    --     tooltip:SetCell(row, 1, format("%sRaging|r %sVolcanic|r %sTyrannical|r", addon:GetKeystoneLevelColour(4), addon:GetKeystoneLevelColour(7), addon:GetKeystoneLevelColour(10)), nil, nil, addon.options.showRecords and 5 or 3)
    -- end
end

local function OnEnter(self)
    if not private.compiled then CompileTooltipInformation() end

    self.tooltip = LibQTip:Acquire("BrokerKeystonesTooltip", addon.options.showRecords and 3 or 2)

    OnTooltipShow(self.tooltip)

    self.tooltip:SmartAnchorTo(self)
    self.tooltip:Show()
end

local function OnLeave(self)
    if self.tooltip then
        LibQTip:Release(self.tooltip)
        self.tooltip = nil
    end
end

addon:RegisterCallback(addon.EVENT_KEYSTONE, ResetTooltipInformation)
addon:RegisterCallback(addon.EVENT_RECORD, ResetTooltipInformation)

addon:RegisterCallback(addon.EVENT_BROKER_INITIALISED, function(_, broker)
    private.broker = broker
    if not addon.options.useClassicTooltip then
        broker.OnEnter = OnEnter
        broker.OnLeave = OnLeave
    end
end)

addon:RegisterCallback(addon.EVENT_OPTION_CHANGED, function(_, key, newValue, oldValue)
    if key ~= private.optionName or not private.broker then return end

    if newValue and private.broker.OnEnter == OnEnter then
        private.broker.OnEnter = nil
        private.broker.OnLeave = nil
        private.broker.OnTooltipShow = nil
    else
        private.broker.OnEnter = OnEnter
        private.broker.OnLeave = OnLeave
    end
end)
